# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import urllib.parse
from typing import Any

from .stringtype import StringType


class HTTPResourceLocator(StringType):
    max_length = 2048

    @classmethod
    def validate(cls, v: str, _: Any = None):
        p = urllib.parse.urlparse(v)
        if p.scheme not in {'http', 'https'}:
            raise ValueError(f"Not a valid URL: {v[:128]}")
        return cls(v)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Literal

import pydantic

from canonical.ext.oauth.types import AuthenticationMethod
from canonical.ext.oauth.types import RedirectURI
from .grantmodel import GrantModel


class AuthorizationCodeRequest(GrantModel):
    grant_type: Literal['authorization_code'] = pydantic.Field(
        default='authorization_code',
        title="Grant Type",
        description=(
            "Value **must** be set to `authorization_code`."
        )
    )

    code: str = pydantic.Field(
        default=...,
        title="Authorization Code",
        description=(
            "The authorization code received from the authorization server."
        )
    )

    redirect_uri: RedirectURI | None = pydantic.Field(
        default=None,
        title="Redirect URI",
        description=(
            "Required, if the `redirect_uri` parameter was included in the "
            "authorization request, and their values **must** be identical."
        )
    )

    client_id: str | None = pydantic.Field(
        default=None,
        title="Client ID",
        description=(
            "Required, if the client is not authenticating with "
            "the authorization serve"
        )
    )

    def identify(self, client_id: str) -> None:
        self.client_id = client_id

    def set_client_secret(self, mode: AuthenticationMethod, secret: str) -> None:
        if mode != 'client_secret_post':
            raise NotImplementedError
        self.client_secret = secret


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import TypeVar

from canonical.ext.api import APIResourceType
from .base import BaseOperation


T = TypeVar('T', bound=APIResourceType)


class Default(BaseOperation[T]):

    def __init__(self, model: type[APIResourceType], **kwargs: Any):
        super().__init__(verb=self.verb, model=model, **kwargs)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

from .base import BaseCache


class NullCache(BaseCache):
    __module__: str = 'canonical.ext.cache'

    async def fetch(self, *args: Any, **kwargs: Any) -> bytes | None:
        return None

    async def put(self, *args: Any, **kwargs: Any) -> None:
        return None


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .authorizationresponse import AuthorizationResponse
from .error import Error
from .tokenresponse import TokenResponse
from .protocolviolation import ProtocolViolation
from .responsemodel import ResponseModel


__all__: list[str] = [
    'AuthorizationResponse',
    'Error',
    'ProtocolViolation',
    'TokenResponse',
    'ResponseModel',
]


# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import ClassVar
from typing import Self

import pydantic
import pydantic.fields


class BaseReference:
    _default_reference_model: ClassVar[type[Any] | None] = None
    _reference_model: type[Any] | None = pydantic.PrivateAttr(default_factory=lambda: None)

    @classmethod
    def __init_subclass__(cls, ref: type | None = None):
        if ref is not None:
            cls._default_reference_model = ref

    @property
    def scalar(self) -> int | str:
        raise NotImplementedError

    def attach_to_namespace(self, namespace: str) -> None:
        raise NotImplementedError

    def as_name(self) -> int | str:
        raise NotImplementedError

    def get_model(self) -> type[Any]:
        model = self._reference_model or self._default_reference_model
        if model is None or isinstance(model, pydantic.fields.ModelPrivateAttr):
            raise TypeError(f"{type(self).__name__} is not attached to a model.")
        return model

    def get_namespace(self) -> str | None:
        raise NotImplementedError

    def has_model(self) -> bool:
        return self._reference_model is not None

    def in_namespace(self, namespace: str) -> bool:
        raise NotImplementedError

    def is_cluster(self) -> bool:
        raise NotImplementedError

    def is_local(self) -> bool:
        raise NotImplementedError

    def is_namespaced(self) -> bool:
        raise NotImplementedError

    def with_model(self, model: type) -> Self:
        self._reference_model = model
        return self

    def with_namespace(self, namespace: str) -> Self:
        raise NotImplementedError


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi

from canonical.ext.fastapi import params
from canonical.ext.fastapi import verbs
from ..clusterclient import ClusterClient
from ..clusterprovider import ClusterProvider
from ..models.requests import AuthorizationRequest
from ..protocols import IRegisteredClient
from ..resources import AuthorizationStateBuilder


class Authorize(verbs.Get[ClusterClient]):
    detail = True
    existing = True
    include_in_schema = True
    path = '/authorize'
    status_code = 302
    verb = 'authorize'

    def get_endpoint_summary(self) -> str:
        return (
            "Authorization endpoint"
        )

    def get_response_description(self) -> str:
        return (
            "Succesful authorization by the resource owner."
        )

    def validate_request(
        self,
        client: IRegisteredClient,
        params: AuthorizationRequest
    ) -> None:
        if not client.allows_redirect(params.redirect_uri):
            self.fail(403, f"Redirection to {params.redirect_uri} is not allowed.")
        if not client.allows_response_type(params.response_type):
            self.fail(403, (
                    "The response_type parameter must match the default response type "
                    "specified by the local client."
                )
            )

    async def handle(
        self,
        http: params.HTTPClient,
        request: fastapi.Request,
        repo: params.ResourceRepository,
        obj: ClusterClient,
        builder: AuthorizationStateBuilder = fastapi.Depends(AuthorizationStateBuilder),
        params: AuthorizationRequest = AuthorizationRequest.as_query()
    ):
        if params.client_id != obj.name:
            self.fail(403, "The client_id parameter must match the client specified in the path.")
        self.validate_request(obj.spec, params)
        provider = await repo.get(
            obj.spec.provider.with_model(ClusterProvider),
            model=ClusterProvider
        )
        meta = await provider.discover(http)
        assert meta.authorization_endpoint is not None
        assert meta.token_endpoint is not None

        builder.with_authorization_endpoint(meta.authorization_endpoint)
        builder.with_token_endpoint(meta.token_endpoint)
        builder.own(obj, controller=True)
        builder.proxy(str(request.url_for('oauth2.callback')))
        builder.with_client(obj.spec)
        builder.with_params(params)
        if not params.redirect_uri:
            builder.with_redirect_uri(obj.spec.default_redirect())
        state = await builder
        await repo.persist(type(state), state)
        return fastapi.responses.RedirectResponse(
            status_code=302,
            url=state.proxy(params)
        )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Protocol

from .itokenrequest import ITokenRequest


class IObtainingClient(Protocol):

    @property
    def id(self) -> str:
        ...

    async def authenticate(self, request: ITokenRequest) -> None:
        ...


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic

from .requestmodel import RequestModel


class GrantModel(RequestModel):
    client_secret: str | None = pydantic.Field(
        default=None,
        title="Client Secret",
        description=(
            "Required if the client is authenticating with the "
            "`client_secret_post` method. The client may omit "
            "the parameter if the client secret is an empty "
            "string"
        )
    )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic

from .formdatamixin import FormDataMixin
from .httprequestmodelmixin import HTTPRequestModelMixin


class FormDataModel(pydantic.BaseModel, HTTPRequestModelMixin, FormDataMixin):

    def __str__(self) -> str:
        assert self.url is not None
        return f'{self.url}?{self.model_dump_urlencoded(mode="query")}'


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import secrets
from typing import Any

from canonical.ext.fastapi.mixins import QueryModelMixin
from canonical.lib import FormDataModel

from ...types import ResponseTypeLiteral
from ...types import RedirectURI


class AuthorizationRequest(FormDataModel, QueryModelMixin):
    model_config = {'extra': 'forbid'}
    client_id: str
    response_type: ResponseTypeLiteral
    redirect_uri: RedirectURI | None = None
    state: str | None = None

    @classmethod
    def new(cls, **kwargs: Any):
        kwargs.setdefault('state', secrets.token_hex(32))
        return cls.model_validate(kwargs)

    def proxy(self, endpoint: str, client_id: str, redirect_uri: str, state: str) -> str:
        """Proxies the authorization request. Return a string containing the
        full authorization URL, with the `client_id`,  `redirect_uri` and
        `state` parameters replaced by the given values.
        """
        clone = self.model_validate({
            **self.model_dump(mode='json'),
            'client_id': client_id,
            'redirect_uri': redirect_uri,
            'state': state
        })
        return str(clone.with_endpoint('GET', endpoint))
        